import { fn, has, symbol as Sym } from '@traversable/registry';
import { minItems } from './items.js';
import { getSchema, isRequired, property, wrapOptional } from './properties.js';
import * as Spec from './specification.js';
import { t } from '@traversable/schema';
export { is, RAW } from './specification.js';
export { Functor, fold, unfold } from './functor.js';
export { isRequired, property, wrapOptional } from './properties.js';
export { minItems } from './items.js';
/** @internal */
const Object_keys = globalThis.Object.keys;
/** @internal */
const isNumber = u => typeof u === 'number';
const getNumericBounds = u => ({
  ...(has('minimum', t.number)(u) && {
    minimum: u.minimum
  }),
  ...(has('maximum', t.number)(u) && {
    maximum: u.maximum
  }),
  ...(has('exclusiveMinimum', t.number)(u) && {
    exclusiveMinimum: u.exclusiveMinimum
  }),
  ...(has('exclusiveMaximum', t.number)(u) && {
    exclusiveMaximum: u.exclusiveMaximum
  })
});
function applyTupleOptionality(xs, {
  min,
  max
}) {
  return min === max ? xs.map(getSchema) : [...xs.slice(0, min).map(getSchema), ...xs.slice(min).map(getSchema)];
}
/* * * * * * * * * * * * * * * * * * *
 *                                   *
 *   no JSON Schema representation   *
 *                                   *
 * * * * * * * * * * * * * * * * * * */
export { Empty as never, Empty as void, Empty as symbol, Empty as undefined, Empty as bigint };
export function empty() {
  return void 0;
}
const Empty = {
  toJsonSchema: empty
};
export function inline(_) {
  return Empty;
}
export { boolean_ as boolean };
const boolean_ = () => Spec.RAW.boolean;
export function integer(schema) {
  const {
    exclusiveMaximum,
    exclusiveMinimum,
    maximum,
    minimum
  } = getNumericBounds(schema);
  let bounds = {};
  if (typeof exclusiveMinimum === 'number') bounds.exclusiveMinimum = exclusiveMinimum;
  if (typeof exclusiveMaximum === 'number') bounds.exclusiveMaximum = exclusiveMaximum;
  if (typeof minimum === 'number') bounds.minimum = minimum;
  if (typeof maximum === 'number') bounds.maximum = maximum;
  return {
    ...Spec.RAW.integer,
    ...bounds
  };
}
export { number_ as number };
function number_(schema) {
  const {
    exclusiveMaximum,
    exclusiveMinimum,
    maximum,
    minimum
  } = getNumericBounds(schema);
  let bounds = {};
  if (typeof exclusiveMinimum === 'number') bounds.exclusiveMinimum = exclusiveMinimum;
  if (typeof exclusiveMaximum === 'number') bounds.exclusiveMaximum = exclusiveMaximum;
  if (typeof minimum === 'number') bounds.minimum = minimum;
  if (typeof maximum === 'number') bounds.maximum = maximum;
  return {
    ...Spec.RAW.number,
    ...bounds
  };
}
export { string_ as string };
function string_(schema) {
  const minLength = has('minLength', isNumber)(schema) ? schema.minLength : null;
  const maxLength = has('maxLength', isNumber)(schema) ? schema.maxLength : null;
  let out = {
    ...Spec.RAW.string
  };
  minLength && void (out.minLength = minLength);
  maxLength && void (out.maxLength = maxLength);
  return out;
}
export function eq(value) {
  return {
    const: value
  };
}
export function optional(x) {
  function toJsonSchema() {
    return getSchema(x);
  }
  toJsonSchema[Sym.optional] = wrapOptional(x);
  return {
    toJsonSchema
  };
}
export function optionalProto(child) {
  optionalProto[Sym.optional] = wrapOptional(child);
  return {
    ...getSchema(child),
    nullable: true
  };
}
//
export function array(child, {
  minLength,
  maxLength
} = {}) {
  return {
    type: 'array',
    items: getSchema(child),
    ...(typeof minLength === 'number' && {
      minLength
    }),
    ...(typeof maxLength === 'number' && {
      maxLength
    })
  };
}
export function record(child) {
  return {
    type: 'object',
    additionalProperties: getSchema(child)
  };
}
export function union(children) {
  return {
    anyOf: children.map(getSchema)
  };
}
export function intersect(children) {
  return {
    allOf: children.map(getSchema)
  };
}
export function tuple(children) {
  const min = minItems(children);
  const max = children.length;
  return {
    type: 'array',
    additionalItems: false,
    items: applyTupleOptionality(children, {
      min,
      max
    }),
    minItems: min,
    maxItems: max
  };
}
export { object_ as object };
function object_(children) {
  const required = Object_keys(children).filter(isRequired(children));
  return {
    type: 'object',
    required,
    properties: fn.map(children, (v, k) => property(required)(v, k))
  };
}
//# sourceMappingURL=jsonSchema.js.map